import random
import string
import unittest

from tuprolog.core import *
from tuprolog.core.operators import DEFAULT_OPERATORS, EMPTY_OPERATORS, STANDARD_OPERATORS, XF, XFY, YF, YFX, FX, FY, \
    XFX, operator, operator_set, OperatorSet, specifier

# noinspection PyUnresolvedReferences
from java.lang import IllegalArgumentException


class AbstractTestOperatorStuff(unittest.TestCase):

    def setUp(self):
        self.priorities = {random.randint(2, 1199) for _ in range(10)}
        self.specifiers = {XF, YF, FX, FY, XFY, YFX, XFX}
        self.functors = {''.join(random.choices(string.ascii_lowercase, k=1)) for _ in range(10)}
        self.random_operators = {operator(functor, spec, priority)
                                 for priority in self.priorities
                                 for spec in self.specifiers
                                 for functor in self.functors}


class TestOperators(AbstractTestOperatorStuff):

    def test_operator_construction(self):
        for priority in self.priorities:
            for spec in self.specifiers:
                for functor in self.functors:
                    op_simple = operator(functor, spec, priority)
                    op_terms = operator(integer(priority), spec.toTerm(), atom(functor))
                    op_struct = operator(struct('op', integer(priority), spec.toTerm(), atom(functor)))

                    for op in {op_simple, op_struct, op_terms}:
                        self.assertEqual(op.getFunctor(), functor)
                        self.assertTrue(op.getSpecifier(), spec)
                        self.assertEqual(op.getPriority(), priority)

                        self.assertEqual(op.functor, functor)
                        self.assertEqual(op.specifier, spec)
                        self.assertEqual(op.priority, priority)

    def test_operator_equality(self):
        for priority in self.priorities:
            for spec in self.specifiers:
                for functor in self.functors:
                    op = operator(functor, spec, priority)
                    op2 = operator(functor, spec, priority)
                    op_terms = operator(integer(priority), spec.toTerm(), atom(functor))
                    op_struct = operator(struct('op', integer(priority), spec.toTerm(), atom(functor)))

                    self.assertTrue(op == op2)
                    self.assertTrue(op == op_terms)
                    self.assertTrue(op == op_struct)

                    different1 = operator(functor + "_", spec, priority)
                    different2 = operator(functor, spec, priority + 1)

                    self.assertTrue(op != different1)
                    self.assertTrue(op != different2)


class TestSpecifiers(AbstractTestOperatorStuff):

    def setUp(self):
        self.actual_specifiers = {XF, YF, FX, FY, XFY, YFX, XFX}
        specifiers_names = {'Xf', 'yF', 'fX', 'fY', 'xFy', 'YfX', 'xFx'}
        self.specifiers_names = {x for name in specifiers_names for x in (name.lower(), name.upper())}
        self.specifiers = {specifier(name) for name in self.specifiers_names}

    def test_specifier_parsing(self):
        # Test specifier is in constants XF, YF, etc.
        for spec in self.specifiers:
            self.assertTrue(spec in self.actual_specifiers)
        self.assertEqual(self.specifiers, self.actual_specifiers)

    def test_specifier_from_terms(self):
        for name in self.specifiers_names:
            self.assertTrue(specifier(atom(name)) in self.actual_specifiers)
            with self.assertRaises(IllegalArgumentException):
                wrong_name = name + 'g'
                specifier(atom(wrong_name))

    def test_specifier_names(self):
        for name in self.specifiers_names:
            spec = specifier(name)
            enum_name = name.upper()
            logic_name = name.lower()
            self.assertEqual(spec.name(), enum_name)
            self.assertEqual(spec.toTerm(), atom(logic_name))
            self.assertEqual(spec.isPrefix(), len(logic_name) == 2 and logic_name[0] == 'f')
            self.assertEqual(spec.isPostfix(), len(logic_name) == 2 and logic_name[-1] == 'f')
            self.assertEqual(spec.isInfix(), len(logic_name) == 3 and logic_name[1] == 'f')


class TestOperatorSets(AbstractTestOperatorStuff):

    def setUp(self):
        super().setUp()
        self.builtin_operator_sets = {EMPTY_OPERATORS, DEFAULT_OPERATORS, STANDARD_OPERATORS,
                                      OperatorSet.CONTROL_FLOW, OperatorSet.ARITHMETIC,
                                      OperatorSet.ARITHMETIC_COMPARISON,
                                      OperatorSet.CLAUSES, OperatorSet.TERM_COMPARISON}
        self.all_builtin_operators = {op for op_set in self.builtin_operator_sets for op in op_set}
        self.builtin_operators_set = operator_set(self.all_builtin_operators)
        self.random_operators_set = operator_set(self.random_operators)
        self.some_operators = [operator('a', XF, 1), operator('b', YF, 2)]

    def test_operator_set_creation(self):
        empty = operator_set()
        self.assertEqual(empty, EMPTY_OPERATORS)
        for op in self.random_operators:
            singleton = operator_set(op)
            self.assertEqual(len(singleton), 1)
            self.assertTrue(op in singleton)
        from_vararg = operator_set(*self.random_operators)
        self.assertEqual(from_vararg, self.random_operators_set)

    def test_operators_set_sum(self):
        my_appended_operators_set = operator_set(self.some_operators[0])
        my_appended_operators_set += self.some_operators[1]
        my_scratch_operators_set = operator_set(self.some_operators)
        self.assertEqual(my_appended_operators_set, my_scratch_operators_set)

    def test_operators_set_sub(self):
        my_scratch_operators_set = operator_set(self.some_operators[0])
        my_subtracted_operators_set = operator_set(self.some_operators)
        my_subtracted_operators_set -= self.some_operators[1]
        self.assertEqual(my_scratch_operators_set, my_subtracted_operators_set)

    def test_operators_set_eq(self):
        my_operator_set_1 = operator_set(self.some_operators)
        my_operator_set_2 = operator_set(self.some_operators)
        self.assertEqual(my_operator_set_1, my_operator_set_2)

    def test_constant_operators_set(self):
        # Assert empty set is properly empty
        self.assertEqual(len(EMPTY_OPERATORS), 0)
        # Assert standard and default sets are equal
        self.assertEqual(DEFAULT_OPERATORS, STANDARD_OPERATORS)
        # Assert standard operators are in set of builtin operators
        for op in STANDARD_OPERATORS:
            self.assertTrue(op in self.all_builtin_operators)


if __name__ == '__main__':
    unittest.main()



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve as _solve
from tuprolog.utils import Taggable
from tuprolog.solve import SolveOptions, solve_options as _solve_options, MAX_TIMEOUT, ALL_SOLUTIONS
from typing import TypeVar, Mapping, Any


ProbExtensions = _solve.ProbExtensions


def probability(taggable: Taggable) -> float:
    return ProbExtensions.getProbability(taggable)


T = TypeVar("T", bound=Taggable, covariant=True)


def set_probability(taggable: T) -> T:
    return ProbExtensions.setProbability(taggable)


def is_probabilistic(solve_opts: SolveOptions) -> bool:
    return ProbExtensions.isProbabilistic(solve_opts)


def set_probabilistic(solve_opts: SolveOptions, value: bool) -> SolveOptions:
    return ProbExtensions.setProbabilistic(solve_opts, value)


def probabilistic(solve_opts: SolveOptions) -> SolveOptions:
    return ProbExtensions.probabilistic(solve_opts)


def solve_options(
        lazy: bool = True,
        timeout: int = MAX_TIMEOUT,
        limit: int = ALL_SOLUTIONS,
        probabilistic: bool = False,
        custom: Mapping[str, Any] = dict(),
        **kwargs: Any
) -> SolveOptions:
    non_probabilistic = _solve_options(lazy, timeout, limit, custom, **kwargs)
    if probabilistic:
        return set_probabilistic(non_probabilistic, True)
    return non_probabilistic


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.plp.*")



# noinspection PyUnresolvedReferences
from typing import Union
import decimal
import jpype.imports
# noinspection PyUnresolvedReferences
import org.gciatto.kt.math as _ktmath
# noinspection PyUnresolvedReferences
import java.lang as _java_lang
from math import ceil

from tuprolog.pyutils import and_then

BigInteger = _ktmath.BigInteger

BigDecimal = _ktmath.BigDecimal

MathContext = _ktmath.MathContext

RoundingMode = _ktmath.RoundingMode

_MAX_LONG = _java_lang.Long.MAX_VALUE

_MIN_LONG = _java_lang.Long.MIN_VALUE

BIG_INTEGER_MAX_LONG = BigInteger.of(_MAX_LONG)

BIG_INTEGER_MIN_LONG = BigInteger.of(_MIN_LONG)

BIG_INTEGER_ZERO = BigInteger.ZERO

BIG_INTEGER_TEN = BigInteger.TEN

BIG_INTEGER_ONE = BigInteger.ONE

BIG_INTEGER_NEGATIVE_ONE = BigInteger.NEGATIVE_ONE

BIG_INTEGER_TWO = BigInteger.TWO

BIG_DECIMAL_ZERO = BigDecimal.ZERO

BIG_DECIMAL_ONE = BigDecimal.ONE

BIG_DECIMAL_ONE_HALF = BigDecimal.ONE_HALF

BIG_DECIMAL_ONE_TENTH = BigDecimal.ONE_TENTH

BIG_DECIMAL_E = BigDecimal.E

BIG_DECIMAL_PI = BigDecimal.PI


def _int_to_bytes(n: int) -> bytes:
    try:
        size = n.bit_length() // 8 + 1
        return n.to_bytes(size, 'big', signed=True)
    except OverflowError as e:
        e.args = ["%s: %d, whose size is %d" % (e.args[0], n, size)]
        raise e


def big_integer(value: Union[str, int], radix: int = None) -> BigInteger:
    if radix is not None:
        assert isinstance(value, str)
        return BigInteger.of(value, radix)
    if isinstance(value, str):
        return BigInteger.of(jpype.JString @ value)
    assert isinstance(value, int)
    bs = _int_to_bytes(value)
    return BigInteger(jpype.JArray(jpype.JByte) @ bs, 0, len(bs))


def python_integer(value: BigInteger) -> int:
    return int.from_bytes(value.toByteArray(), byteorder='big', signed=True)


_python_rounding_modes = {decimal.ROUND_DOWN, decimal.ROUND_HALF_UP, decimal.ROUND_HALF_EVEN, decimal.ROUND_CEILING,
                          decimal.ROUND_FLOOR, decimal.ROUND_UP, decimal.ROUND_HALF_DOWN, decimal.ROUND_05UP}


def jvm_rounding_mode(mode):
    if mode == decimal.ROUND_DOWN:
        return RoundingMode.DOWN
    elif mode == decimal.ROUND_UP:
        return RoundingMode.UP
    elif mode == decimal.ROUND_FLOOR:
        return RoundingMode.FLOOR
    elif mode == decimal.ROUND_CEILING:
        return RoundingMode.CEILING
    elif mode == decimal.ROUND_HALF_UP:
        return RoundingMode.HALF_UP
    elif mode == decimal.ROUND_HALF_EVEN:
        return RoundingMode.HALF_EVEN
    elif mode == decimal.ROUND_HALF_DOWN:
        return RoundingMode.HALF_DOWN
    elif mode == decimal.ROUND_05UP:
        raise ValueError(f"Rounding mode {decimal.ROUND_05UP} has no Java correspondence")
    else:
        raise ValueError(f"Not a rounding mode {mode}")


@and_then(lambda bd: bd.stripTrailingZeros())
def big_decimal(value: Union[str, int, float, decimal.Decimal], precision=0,
                rounding=RoundingMode.HALF_UP) -> BigDecimal:
    if precision is None:
        precision = decimal.getcontext().prec
    assert isinstance(precision, int)
    if rounding is None:
        rounding = jvm_rounding_mode(decimal.getcontext().rounding)
    elif rounding in _python_rounding_modes:
        rounding = jvm_rounding_mode(rounding)
    assert isinstance(rounding, RoundingMode)
    context = MathContext(precision, rounding)
    if isinstance(value, str):
        return BigDecimal.of(value, context)
    if isinstance(value, decimal.Decimal):
        return BigDecimal.of(jpype.JString @ str(value), context)
    if isinstance(value, BigInteger):
        return BigDecimal.of(value, context)
    if isinstance(value, int):
        return BigDecimal.of(big_integer(value), context)
    assert isinstance(value, float)
    return BigDecimal.of(jpype.JDouble @ value, context)


def python_decimal(value: BigDecimal) -> decimal.Decimal:
    return decimal.Decimal(str(value))



import unittest
from tuprolog import *
from tuprolog.core import *

class TestTermCreation(unittest.TestCase):

    def setUp(self):
        self.atom = atom('a')
        self.integer = integer(1)
        self.real = real(1.2)
        self.var = var("X")
        self.true = truth(True)
        self.false = truth(False)
        self.fail = FAIL
        self.struct = struct('f', self.atom, self.integer, self.real)
        self.struct2 = struct('f', [self.atom, self.integer, self.real])


    def test_simple_terms_creation(self):
        self.assertEqual(self.atom.getValue(), 'a')
        self.assertEqual(self.integer.getValue().toInt(), 1)
        self.assertEqual(self.real.getValue().toDouble(), 1.2)
        self.assertEqual(self.var.getName(), 'X')
        self.assertEqual(self.true.getValue(), 'true')
        self.assertEqual(self.false.getValue(), 'false')
        self.assertEqual(self.fail.getValue(), 'fail')

    def test_struct_creation(self):
        self.assertEqual(self.struct.getFunctor(), 'f')
        self.assertEqual(self.struct.getArity(), 3)
        self.assertEqual(self.struct.getArgAt(0), self.atom)
        self.assertEqual(self.struct.getArgAt(1), self.integer)
        self.assertEqual(self.struct.getArgAt(2), self.real)

        self.assertEqual(self.struct, self.struct2)        


if __name__ == '__main__':
    unittest.main()



from tuprolog import logger
import jpype


@jpype.JImplementationFor("it.unibo.tuprolog.core.exception.SubstitutionException")
class _KtSubstitutionException:
    def __jclass_init__(cls):
        pass

    @property
    def substitution(self):
        return self.getSubstitution()


@jpype.JImplementationFor("it.unibo.tuprolog.core.exception.SubstitutionApplicationException")
class _KtSubstitutionApplicationException:
    def __jclass_init__(cls):
        pass

    @property
    def term(self):
        return self.getTerm()


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.core.exception.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.core.parsing as _parsing
from tuprolog.core import Term, Struct, Constant, Var, Atom, Numeric, Integer, Real, Clause
from tuprolog.jvmutils import InputStream
from tuprolog.core.operators import OperatorSet, DEFAULT_OPERATORS
from typing import Union, Iterable
from ._ktadapt import *

TermParser = _parsing.TermParser

TermReader = _parsing.TermReader

ParseException = _parsing.ParseException

InvalidTermTypeException = _parsing.InvalidTermTypeException


def _factory(source, with_default_operators: bool = True, operators: OperatorSet = None):
    if operators is None:
        if with_default_operators:
            return source.getWithDefaultOperators()
        else:
            return source.getWithNoOperator()
    else:
        if with_default_operators:
            return source.withOperators(DEFAULT_OPERATORS.plus(operators))
        else:
            return source.withOperators(operators)


def term_parser(with_default_operators: bool = True, operators: OperatorSet = None) -> TermParser:
    return _factory(TermParser, with_default_operators, operators)


def term_reader(with_default_operators: bool = True, operators: OperatorSet = None) -> TermParser:
    return _factory(TermReader, with_default_operators, operators)


DEFAULT_TERM_PARSER = term_parser()

DEFAULT_TERM_READER = term_reader()


def parse_term(string: str, operators: OperatorSet = None) -> Term:
    return DEFAULT_TERM_PARSER.parse_term(string, operators)


def parse_struct(string: str, operators: OperatorSet = None) -> Struct:
    return DEFAULT_TERM_PARSER.parse_struct(string, operators)


def parse_constant(string: str, operators: OperatorSet = None) -> Constant:
    return DEFAULT_TERM_PARSER.parse_constant(string, operators)


def parse_var(string: str, operators: OperatorSet = None) -> Var:
    return DEFAULT_TERM_PARSER.parse_var(string, operators)


def parse_atom(string: str, operators: OperatorSet = None) -> Atom:
    return DEFAULT_TERM_PARSER.parse_atom(string, operators)


def parse_numeric(string: str, operators: OperatorSet = None) -> Numeric:
    return DEFAULT_TERM_PARSER.parse_numeric(string, operators)


def parse_integer(string: str, operators: OperatorSet = None) -> Integer:
    return DEFAULT_TERM_PARSER.parse_integer(string, operators)


def parse_real(string: str, operators: OperatorSet = None) -> Real:
    return DEFAULT_TERM_PARSER.parse_real(string, operators)


def parse_clause(string: str, operators: OperatorSet = None) -> Clause:
    return DEFAULT_TERM_PARSER.parse_clause(string, operators)


def read_term(input: Union[InputStream, str], operators: OperatorSet = None) -> Term:
    return DEFAULT_TERM_READER.read_term(input, operators)


def read_terms(input: Union[InputStream, str], operators: OperatorSet = None) -> Iterable[Term]:
    return DEFAULT_TERM_READER.read_terms(input, operators)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.core.parsing.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.core.exception as _exceptions
from ._ktadapt import *

TuPrologException = _exceptions.TuPrologException

SubstitutionException = _exceptions.TuPrologException

SubstitutionApplicationException = _exceptions.TuPrologException


logger.debug("Loaded JVM classes from it.unibo.tuprolog.core.exception.*")



from tuprolog import logger
import jpype


@jpype.JImplementationFor("it.unibo.tuprolog.unify.exception.NoUnifyException")
class _KtNoUnifyException:
    def __jclass_init__(cls):
        pass

    @property
    def term1(self):
        return self.getTerm1()

    @property
    def term2(self):
        return self.getTerm2()


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.unify.exception.*")



import unittest
from tuprolog.core import struct, var
from tuprolog.theory.parsing import parse_theory
from tuprolog.solve.plp import solve_options, probability
from tuprolog.solve.problog import problog_solver
from tuprolog.solve.problog.operators import PROBLOG_OPERATORS


class ExemplifyProblog(unittest.TestCase):

    def setUp(self) -> None:
        self.probabilisticTheoryText = """
        0.6::edge(1,2).
        0.1::edge(1,3).
        0.4::edge(2,5).
        0.3::edge(2,6).
        0.3::edge(3,4).
        0.8::edge(4,5).
        0.2::edge(5,6).
        
        path(X,Y) :- edge(X,Y).
        path(X,Y) :- edge(X,Z),Y \\== Z,path(Z,Y).
        """
        self.prints = []

    def print(self, message):
        self.prints.append(message)

    def test_problog(self):
        probabilisticTheory = parse_theory(self.probabilisticTheoryText, PROBLOG_OPERATORS)
        probabilisticSolver = problog_solver(static_kb=probabilisticTheory)
        query = struct('path', var('From'), var('To'))
        for solution in probabilisticSolver.solve(query, solve_options(lazy=True, probabilistic=True)):
            if solution.is_yes:
                self.print(f"yes: {solution.solved_query} with probability {probability(solution)}")

        self.assertEquals(
            self.prints,
            [
                'yes: path(1, 2) with probability 0.6',
                'yes: path(1, 3) with probability 0.1',
                'yes: path(2, 5) with probability 0.4',
                'yes: path(2, 6) with probability 0.356',
                'yes: path(3, 4) with probability 0.3',
                'yes: path(4, 5) with probability 0.8',
                'yes: path(5, 6) with probability 0.2',
                'yes: path(1, 5) with probability 0.25824',
                'yes: path(1, 6) with probability 0.2167296',
                'yes: path(1, 4) with probability 0.03',
                'yes: path(3, 5) with probability 0.24',
                'yes: path(3, 6) with probability 0.04800000000000001',
                'yes: path(4, 6) with probability 0.16000000000000003'
            ]
        )


if __name__ == '__main__':
    unittest.main()



from typing import Union
from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception.error as errors
from tuprolog.core import Term, atom
from tuprolog.solve import ExecutionContext

EvaluationError = errors.EvaluationError

ErrorType = EvaluationError.Type

ERROR_INT_OVERFLOW = ErrorType.INT_OVERFLOW

ERROR_FLOAT_OVERFLOW = ErrorType.FLOAT_OVERFLOW

ERROR_UNDERFLOW = ErrorType.UNDERFLOW

ERROR_ZERO_DIVISOR = ErrorType.ZERO_DIVISOR

ERROR_UNDEFINED = ErrorType.UNDEFINED


def evaluation_error(
        context: ExecutionContext,
        type: ErrorType,
        message: str
) -> EvaluationError:
    return EvaluationError(message, None, context, type, atom(message))


def error_type(name: Union[str, Term]) -> ErrorType:
    if isinstance(name, str):
        return ErrorType.valueOf(name)
    else:
        return ErrorType.fromTerm(name)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.error.EvaluationError.*")



from collections import Mapping
from itertools import chain

from tuprolog import logger

# noinspection PyUnresolvedReferences
import jpype

from tuprolog.pyutils import iterable_or_varargs


@jpype.JImplementationFor("it.unibo.tuprolog.solve.SolverFactory")
class _KtSolverFactory:
    def __jclass_init__(self):
        pass


@jpype.JImplementationFor("it.unibo.tuprolog.solve.SolveOptions")
class _KtSolveOptions:

    _static_keys = {'lazy', 'is_lazy', 'eager', 'is_eager', 'timeout', 'limit'}

    def __jclass_init__(self):
        Mapping.register(self)

    @property
    def is_lazy(self):
        return self.isLazy()

    @property
    def is_eager(self):
        return self.isEager()

    @property
    def timeout(self):
        return self.getTimeout()

    @property
    def limit(self):
        return self.getLimit()

    @property
    def options(self):
        return self.getOptions()

    def __len__(self):
        return 4 + len(self.options)

    def __iter__(self):
        return chain(_KtSolveOptions._static_keys, self.options)

    def __contains__(self, item):
        return item in _KtSolveOptions._static_keys

    def __getitem__(self, item, default=None):
        if item in {'lazy', 'is_lazy'}:
            return self.is_lazy
        elif item in {'eager', 'is_eager'}:
            return self.is_eager
        elif item == 'timeout':
            return self.timeout
        elif item == 'limit':
            return self.limit
        elif item in self.options:
            return self.options[item]
        elif default is not None:
            return default
        return KeyError(f"No such option: {item}")


@jpype.JImplementationFor("it.unibo.tuprolog.solve.ExecutionContextAware")
class _KtExecutionContextAware:
    def __jclass_init__(self):
        pass

    @property
    def libraries(self):
        return self.getLibraries()

    @property
    def flags(self):
        return self.getFlags()

    @property
    def static_kb(self):
        return self.getStaticKb()

    @property
    def dynamic_kb(self):
        return self.getDynamicKb()

    @property
    def operators(self):
        return self.getOperators()

    @property
    def input_channels(self):
        return self.getInputChannles()

    @property
    def output_channels(self):
        return self.getOutputChannles()

    @property
    def standard_output(self):
        return self.getStandardOutput()

    @property
    def standard_input(self):
        return self.getStandardInput()

    @property
    def standard_error(self):
        return self.getStandardError()

    @property
    def warnings(self):
        return self.getWarnings()


@jpype.JImplementationFor("it.unibo.tuprolog.solve.Solver")
class _KtSolver:
    def __jclass_init__(self):
        pass

    @jpype.JOverride
    def solve(self, goal, options=None):
        if options is None:
            return self.solve_(goal)
        elif options.is_eager:
            return self.solveList(goal, options)
        else:
            return self.solve_(goal, options)

    @jpype.JOverride
    def solve_once(self, goal, options=None):
        if options is None:
            return self.solveOnce(goal)
        else:
            return self.solveOnce(goal, options)


@jpype.JImplementationFor("it.unibo.tuprolog.solve.MutableSolver")
class _KtMutableSolver:
    def __jclass_init__(self):
        pass

    def load_library(self, library):
        return self.loadLibrary(library)

    def unload_library(self, library):
        return self.unloadLibrary(library)

    def set_libraries(self, libraries):
        return self.setLibraries(libraries)

    def load_static_kb(self, theory):
        return self.loadStaticKb(theory)

    def load_static_clauses(self, *clauses):
        return iterable_or_varargs(clauses, lambda cs: self.loadStaticClauses(cs))

    def append_static_kb(self, theory):
        return self.appendStaticKb(theory)

    def reset_static_kb(self):
        return self.resetStaticKb()

    def load_dynamic_kb(self, theory):
        return self.loadDynamicKb(theory)

    def load_dynamic_clauses(self, *clauses):
        return iterable_or_varargs(clauses, lambda cs: self.loadDynamicClauses(cs))

    def append_dynamic_kb(self, theory):
        return self.appendDynamicKb(theory)

    def reset_dynamic_kb(self):
        return self.resetDynamicKb()

    def assert_a(self, clause):
        return self.assertA(clause)

    def assert_z(self, clause):
        return self.assertZ(clause)

    def retract_all(self, clause):
        return self.retractAll(clause)

    def set_flag(self, *args):
        return self.setFlag(*args)

    @property
    def standard_output(self):
        return self.getStandardOutput()

    @property
    def standard_input(self):
        return self.getStandardInput()

    @property
    def standard_error(self):
        return self.getStandardError()

    @property
    def warnings(self):
        return self.getWarnings()

    @standard_input.setter
    def standard_input(self, channel):
        return self.setStandardInput(channel)

    @standard_output.setter
    def standard_output(self, channel):
        return self.setStandardOutput(channel)

    @standard_error.setter
    def standard_error(self, channel):
        return self.setStandardError(channel)

    @warnings.setter
    def warnings(self, channel):
        return self.setWarnings(channel)


@jpype.JImplementationFor("it.unibo.tuprolog.solve.Solution")
class _KtSolution:
    def __jclass_init__(self):
        pass

    @property
    def is_yes(self):
        return self.isYes()

    @property
    def is_no(self):
        return self.isNo()

    @property
    def is_halt(self):
        return self.isHalt()

    @property
    def substitution(self):
        return self.getSubstitution()

    @property
    def exception(self):
        return self.getExecption()

    @property
    def solved_query(self):
        return self.getSolvedQuery()

    def clean_up(self):
        return self.cleanUp()

    def value_of(self, variable):
        return self.valueOf(variable)


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.solve.*")



from typing import Union
# noinspection PyUnresolvedReferences
import jpype.imports
from tuprolog import logger
from tuprolog.core import Term
from tuprolog.solve import ExecutionContext, Signature
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception.error as errors


PermissionError = errors.PermissionError

Operation = PermissionError.Operation

Permission = PermissionError.Permission

OPERATION_ACCESS = Operation.ACCESS

OPERATION_ADD_ALIAS = Operation.ADD_ALIAS

OPERATION_CLOSE = Operation.CLOSE

OPERATION_CREATE = Operation.CREATE

OPERATION_INPUT = Operation.INPUT

OPERATION_INVOKE = Operation.INVOKE

OPERATION_MODIFY = Operation.MODIFY

OPERATION_OPEN = Operation.OPEN

OPERATION_OUTPUT = Operation.OUTPUT

OPERATION_REPOSITION = Operation.REPOSITION

PERMISSION_BINARY_STREAM = Permission.BINARY_STREAM

PERMISSION_FLAG = Permission.FLAG

PERMISSION_OPERATOR = Permission.OPERATOR

PERMISSION_PAST_END_OF_STREAM = Permission.PAST_END_OF_STREAM

PERMISSION_PRIVATE_PROCEDURE = Permission.PRIVATE_PROCEDURE

PERMISSION_SOURCE_SINK = Permission.SOURCE_SINK

PERMISSION_STATIC_PROCEDURE = Permission.STATIC_PROCEDURE

PERMISSION_OOP_METHOD = Permission.OOP_METHOD

PERMISSION_STREAM = Permission.STREAM

PERMISSION_TEXT_STREAM = Permission.TEXT_STREAM


def permission_error(
        context: ExecutionContext,
        procedure: Signature,
        operation: Operation,
        permission: Permission,
        culprit: Term
) -> PermissionError:
    return PermissionError.of(context, procedure, operation, permission, culprit)


def operation(name: Union[str, Term]) -> Operation:
    if isinstance(name, str):
        return Operation.of(name)
    else:
        return Operation.fromTerm(name)


def permission(name: Union[str, Term]) -> Permission:
    if isinstance(name, str):
        return Permission.of(name)
    else:
        return Permission.fromTerm(name)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.error.PermissionError.*")



from functools import reduce
from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
from tuprolog.core.operators import OperatorSet, operator_set
from tuprolog.theory import Theory, theory
from tuprolog.solve import Signature
from tuprolog.solve.primitive import Primitive
from tuprolog.solve.function import LogicFunction
from tuprolog.jvmutils import jiterable
from typing import Union, Mapping, Iterable
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.library as _library


Library = _library.Library

Libraries = _library.Libraries

LibraryGroup = _library.LibraryGroup

AliasedLibrary = _library.AliasedLibrary


def library(
        alias: str = None,
        primitives: Mapping[Signature, Primitive] = dict(),
        theory: Theory = theory(),
        operators: OperatorSet = operator_set(),
        functions: Mapping[Signature, LogicFunction] = dict(),
) -> Union[Library, AliasedLibrary]:
    if alias is None:
        return Library.unaliased(primitives, theory, operators, functions)
    else:
        return Library.aliased(alias, primitives, theory, operators, functions)


def aliased(alias: str, library: Library) -> AliasedLibrary:
    return Library.of(alias, library)


def libraries(
        *libs: Union[Libraries, AliasedLibrary, Iterable[Union[Libraries, AliasedLibrary]]],
        **kwargs: Library
) -> Libraries:
    all_libraries = []
    aliased_libs = []
    queue = list(libs)
    while len(queue) > 0:
        current = queue.pop()
        if isinstance(current, Libraries):
            all_libraries.append(current)
        elif isinstance(current, AliasedLibrary):
            aliased_libs.append(current)
        else:
            queue.extend(current)
    for alias in kwargs:
        aliased_libs.append(aliased(alias, kwargs[alias]))
    first = Libraries.of(jiterable(aliased_libs))
    return reduce(lambda a, b: a.plus(b), all_libraries, first)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.library.*")


from typing import Union
# noinspection PyUnresolvedReferences
import jpype.imports

from tuprolog import logger
from tuprolog.core import Term
from tuprolog.solve import ExecutionContext, Signature

# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception.error as errors

RepresentationError = errors.RepresentationError

Limit = RepresentationError.Limit

LIMIT_CHARACTER = Limit.CHARACTER

LIMIT_CHARACTER_CODE = Limit.CHARACTER_CODE

LIMIT_IN_CHARACTER_CODE = Limit.IN_CHARACTER_CODE

LIMIT_MAX_ARITY = Limit.MAX_ARITY

LIMIT_MAX_INTEGER = Limit.MAX_INTEGER

LIMIT_MIN_INTEGER = Limit.MIN_INTEGER

LIMIT_OOP_OBJECT = Limit.OOP_OBJECT

LIMIT_TOO_MANY_VARIABLES = Limit.TOO_MANY_VARIABLES


def representation_error(
        context: ExecutionContext,
        procedure: Signature,
        limit: Limit,
        cause=None
) -> RepresentationError:
    return RepresentationError.of(context, procedure, limit, cause)


def limit(name: Union[str, Term]) -> Limit:
    if isinstance(name, str):
        return Limit.of(name)
    else:
        return Limit.fromTerm(name)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.error.RepresentationError.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.theory.parsing as _parsing
from tuprolog.core import Clause
from tuprolog.core.parsing import _factory
from tuprolog.theory import Theory
from tuprolog.jvmutils import InputStream
from tuprolog.core.operators import OperatorSet
from typing import Union, Iterable
from ._ktadapt import *


ClausesParser = _parsing.ClausesParser

ClausesReader = _parsing.ClausesReader


def clauses_parser(with_default_operators: bool = True, operators: OperatorSet = None) -> ClausesParser:
    return _factory(ClausesParser, with_default_operators, operators)


def clauses_reader(with_default_operators: bool = True, operators: OperatorSet = None) -> ClausesReader:
    return _factory(ClausesReader, with_default_operators, operators)


DEFAULT_CLAUSES_PARSER = clauses_parser()

DEFAULT_CLAUSES_READER = clauses_reader()


def parse_theory(string: str, operators: OperatorSet = None) -> Theory:
    return DEFAULT_CLAUSES_PARSER.parse_theory(string, operators)


def parse_clauses(string: str, operators: OperatorSet = None, lazy: bool = True) -> Iterable[Clause]:
    if lazy:
        return DEFAULT_CLAUSES_PARSER.parse_clauses_lazily(string, operators)
    else:
        return DEFAULT_CLAUSES_PARSER.parse_clauses(string, operators)


def read_theory(input: Union[InputStream, str], operators: OperatorSet = None) -> Theory:
    return DEFAULT_CLAUSES_READER.read_theory(input, operators)


def read_clauses(input: Union[InputStream, str], operators: OperatorSet = None, lazy: bool = True) -> Iterable[Clause]:
    if lazy:
        return DEFAULT_CLAUSES_READER.read_clauses_lazily(input, operators)
    else:
        return DEFAULT_CLAUSES_READER.read_clauses(input, operators)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.theory.parsing.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.stdlib.primitive as _primitive


Abolish = _primitive.Abolish.INSTANCE

Arg = _primitive.Arg.INSTANCE

ArithmeticEqual = _primitive.ArithmeticEqual.INSTANCE

ArithmeticGreaterThan = _primitive.ArithmeticGreaterThan.INSTANCE

ArithmeticGreaterThanOrEqualTo = _primitive.ArithmeticGreaterThanOrEqualTo.INSTANCE

ArithmeticLowerThan = _primitive.ArithmeticLowerThan.INSTANCE

ArithmeticLowerThanOrEqualTo = _primitive.ArithmeticLowerThanOrEqualTo.INSTANCE

ArithmeticNotEqual = _primitive.ArithmeticNotEqual.INSTANCE

Assert = _primitive.Assert.INSTANCE

AssertA = _primitive.AssertA.INSTANCE

AssertZ = _primitive.AssertZ.INSTANCE

Atom = _primitive.Atom.INSTANCE

AtomChars = _primitive.AtomChars.INSTANCE

AtomCodes = _primitive.AtomCodes.INSTANCE

AtomConcat = _primitive.AtomConcat.INSTANCE

AtomLength = _primitive.AtomLength.INSTANCE

Atomic = _primitive.Atomic.INSTANCE

BagOf = _primitive.BagOf.INSTANCE

Between = _primitive.Between.INSTANCE

Callable = _primitive.Callable.INSTANCE

CharCode = _primitive.CharCode.INSTANCE

Clause = _primitive.Clause.INSTANCE

Compound = _primitive.Compound.INSTANCE

CopyTerm = _primitive.CopyTerm.INSTANCE

CurrentFlag = _primitive.CurrentFlag.INSTANCE

CurrentOp = _primitive.CurrentOp.INSTANCE

EnsureExecutable = _primitive.EnsureExecutable.INSTANCE

FindAll = _primitive.FindAll.INSTANCE

Float = _primitive.Float.INSTANCE

Functor = _primitive.Functor.INSTANCE

GetDurable = _primitive.GetDurable.INSTANCE

GetEphemeral = _primitive.GetEphemeral.INSTANCE

GetPersistent = _primitive.GetPersistent.INSTANCE

Ground = _primitive.Ground.INSTANCE

Halt = _primitive.Halt.INSTANCE

Halt1 = _primitive.Halt1.INSTANCE

Integer = _primitive.Integer.INSTANCE

Is = _primitive.Is.INSTANCE

Natural = _primitive.Natural.INSTANCE

NewLine = _primitive.NewLine.INSTANCE

NonVar = _primitive.NonVar.INSTANCE

NotUnifiableWith = _primitive.NotUnifiableWith.INSTANCE

Number = _primitive.Number.INSTANCE

NumberChars = _primitive.NumberChars.INSTANCE

NumberCodes = _primitive.NumberCodes.INSTANCE

Op = _primitive.Op.INSTANCE

Repeat = _primitive.Repeat.INSTANCE

Retract = _primitive.Retract.INSTANCE

RetractAll = _primitive.RetractAll.INSTANCE

Reverse = _primitive.Reverse.INSTANCE

SetDurable = _primitive.SetDurable.INSTANCE

SetEphemeral = _primitive.SetEphemeral.INSTANCE

SetFlag = _primitive.SetFlag.INSTANCE

SetOf = _primitive.SetOf.INSTANCE

SetPersistent = _primitive.SetPersistent.INSTANCE

Sleep = _primitive.Sleep.INSTANCE

SubAtom = _primitive.SubAtom.INSTANCE

TermGreaterThan = _primitive.TermGreaterThan.INSTANCE

TermGreaterThanOrEqualTo = _primitive.TermGreaterThanOrEqualTo.INSTANCE

TermIdentical = _primitive.TermIdentical.INSTANCE

TermLowerThan = _primitive.TermLowerThan.INSTANCE

TermLowerThanOrEqualTo = _primitive.TermLowerThanOrEqualTo.INSTANCE

TermNotIdentical = _primitive.TermNotIdentical.INSTANCE

TermNotSame = _primitive.TermNotSame.INSTANCE

TermSame = _primitive.TermSame.INSTANCE

UnifiesWith = _primitive.UnifiesWith.INSTANCE

Univ = _primitive.Univ.INSTANCE

Var = _primitive.Var.INSTANCE

Write = _primitive.Write.INSTANCE


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.stdlib.primitive.*")



from typing import Iterable, Union
from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception.error as errors
from tuprolog.core import Term
from tuprolog.solve import ExecutionContext, Signature

DomainError = errors.DomainError

Domain = DomainError.Expected

DOMAIN_wATOM_PROPERTY = Domain.ATOM_PROPERTY

DOMAIN_BUFFERING_MODE = Domain.BUFFERING_MODE

DOMAIN_CHARACTER_CODE_LIST = Domain.CHARACTER_CODE_LIST

DOMAIN_CLOSE_OPTION = Domain.CLOSE_OPTION

DOMAIN_DATE_TIME = Domain.DATE_TIME

DOMAIN_EOF_ACTION = Domain.EOF_ACTION

DOMAIN_FLAG_VALUE = Domain.FLAG_VALUE

DOMAIN_FORMAT_CONTROL_SEQUENCE = Domain.FORMAT_CONTROL_SEQUENCE

DOMAIN_IO_MODE = Domain.IO_MODE

DOMAIN_WELL_FORMED_LIST = Domain.WELL_FORMED_LIST

DOMAIN_NON_EMPTY_LIST = Domain.NON_EMPTY_LIST

DOMAIN_NOT_LESS_THAN_ZERO = Domain.NOT_LESS_THAN_ZERO

DOMAIN_OPERATOR_PRIORITY = Domain.OPERATOR_PRIORITY

DOMAIN_OPERATOR_SPECIFIER = Domain.OPERATOR_SPECIFIER

DOMAIN_ORDER = Domain.ORDER

DOMAIN_OS_FILE_PERMISSION = Domain.OS_FILE_PERMISSION

DOMAIN_OS_FILE_PROPERTY = Domain.OS_FILE_PROPERTY

DOMAIN_OS_PATH = Domain.OS_PATH

DOMAIN_PREDICATE_PROPERTY = Domain.PREDICATE_PROPERTY

DOMAIN_FLAG = Domain.FLAG

DOMAIN_READ_OPTION = Domain.READ_OPTION

DOMAIN_SELECTABLE_ITEM = Domain.SELECTABLE_ITEM

DOMAIN_SOCKET_ADDRESS = Domain.SOCKET_ADDRESS

DOMAIN_SOCKET_DOMAIN = Domain.SOCKET_DOMAIN

DOMAIN_SOURCE_SINK = Domain.SOURCE_SINK

DOMAIN_STREAM = Domain.STREAM

DOMAIN_STREAM_OPTION = Domain.STREAM_OPTION

DOMAIN_STREAM_OR_ALIAS = Domain.STREAM_OR_ALIAS

DOMAIN_STREAM_POSITION = Domain.STREAM_POSITION

DOMAIN_STREAM_PROPERTY = Domain.STREAM_PROPERTY

DOMAIN_STREAM_SEEK_METHOD = Domain.STREAM_SEEK_METHOD

DOMAIN_STREAM_TYPE = Domain.STREAM_TYPE

DOMAIN_TERM_STREAM_OR_ALIAS = Domain.TERM_STREAM_OR_ALIAS

DOMAIN_VAR_BINDING_OPTION = Domain.VAR_BINDING_OPTION

DOMAIN_WRITE_OPTION = Domain.WRITE_OPTION

DOMAIN_CLAUSE = Domain.CLAUSE

DOMAIN_RULE = Domain.RULE

DOMAIN_FACT = Domain.FACT

DIRECTIVE = Domain.DIRECTIVE


def domain_error_for_flag_values(
        context: ExecutionContext,
        procedure: Signature,
        flag_values: Iterable[Term],
        actual: Term,
        index: int = None
) -> DomainError:
    return DomainError.forFlagValues(context, procedure, flag_values, actual, index)


def domain_error_for_argument(
        context: ExecutionContext,
        procedure: Signature,
        expected: Domain,
        actual: Term,
        index: int = None
) -> DomainError:
    return DomainError.forArgument(context, procedure, expected, actual, index)


def domain_error_for_term(
        context: ExecutionContext,
        expected: Domain,
        actual_value: Term,
) -> DomainError:
    return DomainError.forTerm(context, expected, actual_value)


def domain_error_for_goal(
        context: ExecutionContext,
        procedure: Signature,
        expected: Domain,
        actual: Term,
) -> DomainError:
    return DomainError.forGoal(context, procedure, expected, actual)


def domain(name: Union[str, Term]) -> Domain:
    if isinstance(name, str):
        return Domain.of(name)
    else:
        return Domain.fromTerm(name)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.error.DomainError.*")



from typing import Union
# noinspection PyUnresolvedReferences
import jpype.imports
from tuprolog import logger
from tuprolog.core import Term
from tuprolog.solve import ExecutionContext, Signature
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception.error as errors


TypeError = errors.TypeError

Type = TypeError.Expected

TYPE_ATOM = Type.ATOM

TYPE_ATOMIC = Type.ATOMIC

TYPE_BOOLEAN = Type.BOOLEAN

TYPE_BYTE = Type.BYTE

TYPE_CALLABLE = Type.CALLABLE

TYPE_CHARACTER = Type.CHARACTER

TYPE_COMPOUND = Type.COMPOUND

TYPE_DEALIASING_EXPRESSION = Type.DEALIASING_EXPRESSION

TYPE_EVALUABLE = Type.EVALUABLE

TYPE_FLOAT = Type.FLOAT

TYPE_INTEGER = Type.INTEGER

TYPE_IN_CHARACTER = Type.IN_CHARACTER

TYPE_LIST = Type.LIST

TYPE_NUMBER = Type.NUMBER

TYPE_OBJECT_REFERENCE = Type.OBJECT_REFERENCE

TYPE_PAIR = Type.PAIR

TYPE_PREDICATE_INDICATOR = Type.PREDICATE_INDICATOR

TYPE_REFERENCE = Type.REFERENCE

TYPE_TYPE_REFERENCE = Type.TYPE_REFERENCE

TYPE_URL = Type.URL

TYPE_VARIABLE = Type.VARIABLE

def type_error_for_flag_values(
        context: ExecutionContext,
        procedure: Signature,
        expected: Type,
        actual: Term,
        message: str
) -> TypeError:
    return TypeError.of(context, procedure, expected, actual, message)


def type_error_for_argument_list(
        context: ExecutionContext,
        procedure: Signature,
        expected: Type,
        actual: Term,
        index: int = None
) -> TypeError:
    return TypeError.forArgumentList(context, procedure, expected, actual, index)


def type_error_for_argument(
        context: ExecutionContext,
        procedure: Signature,
        expected: Type,
        actual: Term,
        index: int = None
) -> TypeError:
    return TypeError.forArgument(context, procedure, expected, actual, index)


def type_error_for_term(
        context: ExecutionContext,
        expected: Type,
        actual_value: Term,
) -> TypeError:
    return TypeError.forTerm(context, expected, actual_value)


def type_error_for_goal(
        context: ExecutionContext,
        procedure: Signature,
        expected: Type,
        actual: Term,
) -> TypeError:
    return TypeError.forGoal(context, procedure, expected, actual)


def type(name: Union[str, Term]) -> Type:
    if isinstance(name, str):
        return TypeError.valueOf(name)
    else:
        return TypeError.fromTerm(name)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.error.TypeError.*")



from tuprolog import logger
import jpype


@jpype.JImplementationFor("it.unibo.tuprolog.utils.Taggable")
class _KtTaggable:
    def __jclass_init__(cls):
        pass

    @property
    def tags(self):
        return self.getTags()

    def get_tag(self, name):
        return self.getTag(name)

    def replace_tags(self, tags):
        return self.replaceTags(tags)

    def contains_tag(self, name):
        return self.containsTag(name)


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.utils.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.problog as _problog

Operators = _problog.Operators

ANNOTATION_OPERATOR = Operators.ANNOTATION_OPERATOR

PROBLOG_SPECIFIC_OPERATORS = Operators.PROBLOG_SPECIFIC_OPERATORS

PROBLOG_OPERATORS = Operators.PROBLOG_OPERATORS

logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.problog.operators")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.stdlib as _stdlib


CommonBuiltins = _stdlib.CommonBuiltins

CommonFunctions = _stdlib.CommonFunctions

CommonPrimitives = _stdlib.CommonPrimitives

CommonRules = _stdlib.CommonRules


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.stdlib.*")



from tuprolog import logger
import jpype

from tuprolog.jvmioutils import ensure_input_steam


@jpype.JImplementationFor("it.unibo.tuprolog.theory.parsing.ClausesParser")
class _KtClausesParser:
    def __jclass_init__(cls):
        pass

    @property
    def default_operator_set(self):
        return self.getDefaultOperatorSet()

    def _parse(self, method, input, operators):
        if operators is None:
            return method(input)
        else:
            return method(input, operators)

    def parse_theory(self, input, operators):
        return self._parse(self.parseTheory, input, operators)

    def parse_clauses_lazily(self, input, operators):
        return self._parse(self.parseClausesLazily, input, operators)

    def parse_clauses(self, input, operators):
        return list(self.parse_clauses_lazily(input, operators))



@jpype.JImplementationFor("it.unibo.tuprolog.theory.parsing.ClausesReader")
class _KtClausesReader:
    def __jclass_init__(cls):
        pass

    @property
    def default_operator_set(self):
        return self.getDefaultOperatorSet()

    def _read(self, method, input, operators):
        input_stream = ensure_input_steam(input)
        if operators is None:
            return method(input_stream)
        else:
            return method(input_stream, operators)

    def read_theory(self, input, operators):
        return self._read(self.readTheory, input, operators)

    def read_clauses_lazily(self, input, operators):
        return self._read(self.readClausesLazily, input, operators)

    def read_clauses(self, input, operators):
        return list(self.read_clauses_lazily(input, operators))


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.theory.parsing.*")



from decimal import Decimal

from tuprolog import logger
import jpype
import jpype.imports
from ._ktadapt import *
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.core as _core
from tuprolog.pyutils import iterable_or_varargs
from tuprolog.jvmutils import jiterable, jmap
from typing import Iterable, Dict, Tuple as PyTuple
from ._ktmath import *

Atom = _core.Atom

Block = _core.Block

Clause = _core.Clause

Cons = _core.Cons

Constant = _core.Constant

Directive = _core.Directive

Empty = _core.Empty

EmptyBlock = _core.EmptyBlock

EmptyList = _core.EmptyList

Fact = _core.Fact

Formatter = _core.Formatter

Indicator = _core.Indicator

Integer = _core.Integer

List = _core.List

Numeric = _core.Numeric

Real = _core.Real

Recursive = _core.Recursive

Rule = _core.Rule

Scope = _core.Scope

Struct = _core.Struct

Substitution = _core.Substitution

Term = _core.Term

TermComparator = _core.TermComparator

TermConvertible = _core.TermConvertible

TermFormatter = _core.TermFormatter

Terms = _core.Terms

TermVisitor = _core.TermVisitor

Truth = _core.Truth

Tuple = _core.Tuple

Var = _core.Var


@jpype.JImplements(TermConvertible)
class AbstractTermConvertible(object):
    @jpype.JOverride
    def toTerm(self):
        raise NotImplementedError()


def atom(string: str) -> Atom:
    return Atom.of(string)


def block(*terms: Union[Term, Iterable[Term]]) -> Block:
    return iterable_or_varargs(terms, lambda ts: Block.of(jiterable(ts)))


def clause(head: Term = None, *body: Union[Term, Iterable[Term]]):
    return iterable_or_varargs(body, lambda bs: Clause.of(head, jiterable(bs)))


def empty_logic_list() -> EmptyList:
    return EmptyList.getInstance()


def cons(head: Term, tail: Term = None) -> Cons:
    if tail is None:
        return Cons.singleton(head)
    else:
        return Cons.of(head, tail)


def directive(*goals: Union[Term, Iterable[Term]]) -> Directive:
    return iterable_or_varargs(goals, lambda gs: Directive.of(jiterable(gs)))


def empty_block() -> Block:
    return EmptyBlock.getInstance()


def fact(struct: Union[Term, Iterable[Term]]) -> Fact:
    return Fact.of(struct)


def indicator(name: Union[str, Term], arity: Union[int, Term]) -> Indicator:
    return Indicator.of(name, arity)


def integer(value: Union[int, BigInteger, str]) -> Integer:
    return Integer.of(big_integer(value))


def real(value: Union[float, BigDecimal, str, Decimal]) -> Real:
    return Real.of(big_decimal(value))


def numeric(value: Union[int, BigInteger, BigDecimal, str, float, Decimal]) -> Numeric:
    if isinstance(value, str):
        return Numeric.of(jpype.JString @ value)
    if isinstance(value, int) or isinstance(value, BigInteger):
        return integer(value)
    return real(value)


def rule(head: Struct, *body: Union[Term, Iterable[Term]]) -> Rule:
    return iterable_or_varargs(body, lambda bs: Rule.of(head, jiterable(bs)))


def struct(functor: str, *args: Union[Term, Iterable[Term]]) -> Struct:
    return iterable_or_varargs(args, lambda xs: Struct.of(functor, jiterable(xs)))


def truth(boolean: bool) -> Truth:
    return Truth.of(boolean)


TRUE = Truth.TRUE

FALSE = Truth.FALSE

FAIL = Truth.FAIL


def logic_list(*items: Union[Term, Iterable[Term]]) -> List:
    return iterable_or_varargs(items, lambda xs: List.of(jiterable(xs)))


def logic_list_from(items: Iterable[Term], last: Term = None) -> List:
    return List.from_(jiterable(items), last)


def logic_tuple(first: Term, second: Term, *others: Union[Term, Iterable[Term]]) -> Tuple:
    return iterable_or_varargs(others, lambda os: Tuple.of(jiterable([first, second] + list(os))))


def var(name: str) -> Var:
    return Var.of(name)


def unifier(assignments: Dict[Var, Term] = {}) -> Substitution.Unifier:
    return Substitution.unifier(jmap(assignments))


def substitution(assignments: Dict[Var, Term] = {}) -> Substitution:
    return Substitution.of(jmap(assignments))


def failed() -> Substitution.Fail:
    return Substitution.failed()


EMPTY_UNIFIER: Substitution.Unifier = substitution()

FAILED_SUBSTITUTION: Substitution.Fail = failed()


def scope(*variables: Union[Var, str]) -> Scope:
    if len(variables) == 0:
        return Scope.empty()
    vars = [var(v) if isinstance(v, str) else v for v in variables]
    return Scope.of(jpype.JArray(Var) @ vars)


def variables(*names: str) -> PyTuple[Var]:
    assert len(names) > 0
    return tuple((var(n) for n in names))


logger.debug("Loaded JVM classes from it.unibo.tuprolog.core.*")



from typing import Union
from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception.error as errors
from tuprolog.core import Term, Atom
from tuprolog.solve import ExecutionContext, Signature

ExistenceError = errors.ExistenceError

ObjectType = ExistenceError.ObjectType

OBJECT_PROCEDURE = ObjectType.PROCEDURE

OBJECT_SOURCE_SINK = ObjectType.SOURCE_SINK

OBJECT_RESOURCE = ObjectType.RESOURCE

OBJECT_STREAM = ObjectType.STREAM

OBJECT_OOP_ALIAS = ObjectType.OOP_ALIAS

OBJECT_OOP_METHOD = ObjectType.OOP_METHOD

OBJECT_OOP_CONSTRUCTOR = ObjectType.OOP_CONSTRUCTOR

OBJECT_OOP_PROPERTY = ObjectType.OOP_PROPERTY


def existence_error(
        context: ExecutionContext,
        type: ObjectType,
        culprit: Term,
        message: str
) -> ExistenceError:
    return ExistenceError.of(context, type, culprit, message)


def existence_error_for_source_sink(
        context: ExecutionContext,
        alias: Union[Atom, str]
) -> ExistenceError:
    return ExistenceError.forSourceSink(context, alias)


def existence_error_for_procedure(
        context: ExecutionContext,
        procedure: Signature
) -> ExistenceError:
    return ExistenceError.forProcedure(context, procedure)


def existence_error_for_stream(
        context: ExecutionContext,
        stream: Term
) -> ExistenceError:
    return ExistenceError.forStream(context, stream)


def existence_error_for_resource(
        context: ExecutionContext,
        name: str
) -> ExistenceError:
    return ExistenceError.forResource(context, name)


def object_type(name: Union[str, Term]) -> ObjectType:
    if isinstance(name, str):
        return ObjectType.of(name)
    else:
        return ObjectType.fromTerm(name)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.error.ExistenceError.*")



from typing import Iterable, Mapping, TypeVar, Union, Callable

T = TypeVar('T')
R = TypeVar('R')


def iterable_or_varargs(
        args: Union[Iterable[T], Iterable[Iterable[T]]],
        dispatch: Callable[[Iterable[T]], R] = lambda x: x
) -> R:
    assert isinstance(args, Iterable)
    if len(args) == 1:
        item = args[0]
        if isinstance(item, Iterable):
            return dispatch(item)
        else:
            return dispatch([item])
    else:
        return dispatch(args)


def dict_or_keyword_args(
        dictionary: Mapping[str, T],
        kwargs: Mapping[str, T],
        dispatch: Callable[[Mapping[str, T]], R] = lambda x: x
) -> R:
    all_data = dict()
    for k in dictionary:
        all_data[k] = dictionary[k]
    for k in kwargs:
        all_data[k] = kwargs[k]
    return dispatch(all_data)


def and_then(continuation):
    def call_and_then_continue(function):
        def wrapper(*args, **kwargs):
            result = function(*args, **kwargs)
            return continuation(result)
        return wrapper
    return call_and_then_continue


def apply_to_result(consumer):
    def call_and_then_apply(function):
        def wrapper(*args, **kwargs):
            result = function(*args, **kwargs)
            consumer(result)
            return result
        return wrapper
    return call_and_then_apply



from decimal import Decimal

from tuprolog import logger
import jpype

from ._ktmath import big_integer, big_decimal, BigInteger, BigDecimal
from tuprolog.utils import *
from typing import Sized, Callable
from tuprolog.jvmutils import jiterable
from tuprolog.pyutils import iterable_or_varargs
from tuprolog.jvmutils import kfunction


@jpype.JImplementationFor("it.unibo.tuprolog.core.Term")
class _KtTerm:
    def __jclass_init__(cls):
        pass

    def __getitem__(self, item, *items):
        return self.get(item, *items)

    @property
    def variables(self):
        return self.getVariables()

    def structurally_equals(self, other):
        return self.structurallyEquals(other)

    @jpype.JOverride()
    def equals(self, other, use_var_complete_name: bool = True):
        return self.equals_(other, use_var_complete_name)

    @property
    def is_var(self):
        return self.isVar()

    @property
    def is_ground(self):
        return self.isGround()

    @property
    def is_struct(self):
        return self.isStruct()

    @property
    def is_truth(self):
        return self.isTruth()

    @property
    def is_recursive(self):
        return self.isRecursive()

    @property
    def is_atom(self):
        return self.isAtom()

    @property
    def is_constant(self):
        return self.isConstant()

    @property
    def is_number(self):
        return self.isNumber()

    @property
    def is_integer(self):
        return self.is_integer()

    @property
    def is_real(self):
        return self.isReal()

    @property
    def is_list(self):
        return self.isList()

    @property
    def is_tuple(self):
        return self.isTuple()

    @property
    def is_block(self):
        return self.isBlock()

    @property
    def is_clause(self):
        return self.isClause()

    @property
    def is_rule(self):
        return self.isRule()

    @property
    def is_fact(self):
        return self.isFact()

    @property
    def is_directive(self):
        return self.isDirective()

    @property
    def is_cons(self):
        return self.isCons()

    @property
    def is_empty_list(self):
        return self.isEmptyList()

    @property
    def is_true(self):
        return self.isTrue()

    @property
    def is_fail(self):
        return self.isFail()

    @property
    def is_indicator(self):
        return self.isIndicator()

    def fresh_copy(self, scope=None):
        if scope is None:
            return self.freshCopy()
        else:
            return self.freshCopy(scope)


@jpype.JImplementationFor("it.unibo.tuprolog.core.Struct")
class _KtStruct:
    def __jclass_init__(cls):
        pass

    @property
    def functor(self):
        return self.getFunctor()

    @property
    def args(self):
        return self.getArgs()

    @property
    def argsSequence(self):
        return self.getArgsSequence()

    @property
    def arity(self):
        return self.getArity()

    @property
    def is_functor_well_formed(self):
        return self.isFunctorWellFormed()

    @property
    def indicator(self):
        return self.getIndicator()

    def get_arg_at(self, index):
        return self.getArgAt(index)

    def add_last(self, argument):
        return self.addLast(argument)

    def add_first(self, argument):
        return self.addFirst(argument)

    def insert_at(self, index, argument):
        return self.addFirst(index, argument)

    def set_functor(self, functor):
        return self.setFunctor(functor)

    def set_args(self, *args):
        return iterable_or_varargs(args, lambda xs: self.setArgs(jiterable(args)))


@jpype.JImplementationFor("it.unibo.tuprolog.core.Var")
class _KtVar:
    def __jclass_init__(cls):
        pass

    @property
    def is_anonymous(self):
        return self.isAnonymous()

    @property
    def name(self):
        return self.getName()

    @property
    def id(self):
        return self.getId()

    @property
    def complete_name(self):
        return self.getCompleteName()

    @property
    def is_name_well_formed(self):
        return self.isNameWellFormed()


@jpype.JImplementationFor("it.unibo.tuprolog.core.Constant")
class _KtConstant:
    def __jclass_init__(cls):
        pass

    @property
    def value(self):
        return self.getValue()


@jpype.JImplementationFor("it.unibo.tuprolog.core.Numeric")
class _KtNumeric:
    def __jclass_init__(cls):
        pass

    @property
    def int_value(self):
        return self.getIntValue()

    @property
    def decimal_value(self):
        return self.getDecimalValue()


@jpype.JImplementationFor("it.unibo.tuprolog.core.Recursive")
class _KtRecursive:
    def __jclass_init__(cls):
        Sized.register(cls)

    @property
    def lazily_unfolded(self):
        return self.getUnfoldedSequence()

    @property
    def unfolded(self):
        return self.getUnfoldedList()

    @property
    def __len__(self):
        return self.getSize()

    def to_iterable(self):
        return self.toSequence()

    def to_list(self):
        return self.toList()

    def to_array(self):
        return self.toArray()


@jpype.JImplementationFor("it.unibo.tuprolog.core.List")
class _KtList:
    def __jclass_init__(cls):
        pass

    @property
    def is_well_formed(self):
        return self.isWellFormed()

    @property
    def last(self):
        return self.getLast()

    def estimated_length(self):
        return self.getEstimatedLength()


@jpype.JImplementationFor("it.unibo.tuprolog.core.Cons")
class _KtCons:
    def __jclass_init__(cls):
        pass

    @property
    def head(self):
        return self.getHead()

    @property
    def tail(self):
        return self.getTail()


@jpype.JImplementationFor("it.unibo.tuprolog.core.Tuple")
class _KtTuple:
    def __jclass_init__(cls):
        pass

    @property
    def left(self):
        return self.getLeft()

    @property
    def right(self):
        return self.getRight()


@jpype.JImplementationFor("it.unibo.tuprolog.core.Clause")
class _KtClause:
    def __jclass_init__(cls):
        pass

    @property
    def head(self):
        return self.getHead()

    @property
    def body(self):
        return self.getBody()

    @property
    def is_well_formed(self):
        return self.isWellFormed()

    @property
    def body_items(self):
        return self.getBodyItems()

    @property
    def body_size(self):
        return self.getBodySize()

    def get_body_item(self, index):
        return self.getBodyItem(index)

    def set_head(self, head):
        return self.setHead(head)

    def set_body(self, term):
        return self.setBody(term)

    def set_head_functor(self, functor):
        return self.setHeadFunctor(functor)

    def set_head_args(self, *args):
        return iterable_or_varargs(args, lambda xs: self.setHeadArgs(jiterable(args)))

    def insert_head_arg(self, index, argument):
        return self.setHeadArg(index, argument)

    def add_first_head_arg(self, argument):
        return self.addFirstHeadArg(argument)

    def add_last_head_arg(self, argument):
        return self.addLastHeadArg(argument)

    def append_head_arg(self, argument):
        return self.appendHeadArg(argument)

    def set_body_items(self, *args):
        return iterable_or_varargs(args, lambda xs: self.setBodyItems(jiterable(args)))

    def insert_body_item(self, index, item):
        return self.insertBodyItem(index, item)

    def add_first_body_item(self, item):
        return self.addFirstBodyItem(item)

    def add_last_body_item(self, item):
        return self.addLastBodyItem(item)

    def append_body_item(self, item):
        return self.appendBodyItem(item)


@jpype.JImplementationFor("it.unibo.tuprolog.core.Rule")
class _KtRule:
    def __jclass_init__(cls):
        pass

    @property
    def head_args(self):
        return self.getHeadArgs()

    @property
    def head_arity(self):
        return self.getHeadArity()

    def get_head_arg(self, index):
        return self.getHeadArg(index)


@jpype.JImplementationFor("it.unibo.tuprolog.core.TermConvertible")
class _KtTermConvertible:
    def __jclass_init__(cls):
        pass

    def to_term(self):
        return self.toTerm()


@jpype.JImplementationFor("it.unibo.tuprolog.core.Substitution")
class _KtSubstitution:
    def __jclass_init__(cls):
        pass

    @property
    def is_success(self):
        return self.isSuccess()

    @property
    def is_failed(self):
        return self.isFailed()

    def apply_to(self, term):
        return self.applyTo(term)

    def get_original(self, variable):
        return self.getOriginal(variable)

    def get_by_name(self, name):
        return self.getByName(name)

    def __add__(self, other):
        return self.plus(other)

    def __sub__(self, other):
        return self.minus(other)

    @jpype.JOverride
    def filter(self, filter):
        if isinstance(filter, Callable):
            return self.filter_(kfunction(1)(filter))
        else:
            return self.filter_(filter)


@jpype.JImplementationFor("it.unibo.tuprolog.core.Scope")
class _KtScope:
    def __jclass_init__(cls):
        pass

    def __contains__(self, item):
        return self.contains(item)

    def __getitem__(self, item):
        return self.get(item)

    @property
    def variables(self):
        return self.getVariables()

    @property
    def fail(self):
        return self.getFail()

    @property
    def empty_list(self):
        return self.getEmptyList()

    @property
    def empty_block(self):
        return self.getEmptyBlock()

    def atom(self, string):
        return self.atomOf(string)

    def block(self, *terms):
        return iterable_or_varargs(terms, lambda ts: self.blockOf(jiterable(ts)))

    def clause(self, head=None, *body):
        return iterable_or_varargs(body, lambda bs: self.clauseOf(head, jiterable(bs)))

    def cons(self, head, tail=None):
        if tail is None:
            return self.listOf(head)
        else:
            return self.consOf(head, tail)

    def directive(self, *goals):
        return iterable_or_varargs(goals, lambda gs: self.directiveOf(jiterable(gs)))

    def fact(self, head):
        return self.factOf(head)

    def indicator(self, name, arity):
        return self.indicatorOf(name, arity)

    def integer(self, value):
        return self.intOf(big_integer(value))

    def real(self, value):
        return self.intOf(big_integer(value))

    def numeric(self, value):
        if isinstance(value, str):
            return self.numOf(value)
        if isinstance(value, int) or isinstance(value, BigInteger):
            return self.intOf(value)
        return self.realOf(value)

    def rule(self, head, *body):
        return iterable_or_varargs(body, lambda bs: self.ruleOf(head, jiterable(bs)))

    def struct(self, functor, *args):
        return iterable_or_varargs(args, lambda xs: self.structOf(functor, jiterable(xs)))

    def truth(self, value):
        return self.truthOf(value)

    def var(self, name):
        return self.varOf(name)

    def list(self, *items):
        return iterable_or_varargs(items, lambda xs: self.listOf(jiterable(xs)))

    def list_from(self, items, last=None):
        return self.listFrom(jiterable(items), last)

    def tuple(self, first, second, *others):
        return iterable_or_varargs(others, lambda os: self.tupleOf(jiterable([first, second] + list(os))))

    @property
    def anonymous(self):
        return self.getAnonymous()

    @property
    def whatever(self):
        return self.getWhatever()


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.core.*")



from tuprolog import logger
import jpype

from tuprolog.jvmioutils import ensure_input_steam


@jpype.JImplementationFor("it.unibo.tuprolog.core.parsing.TermParser")
class _KtTermParser:
    def __jclass_init__(cls):
        pass

    @property
    def default_operator_set(self):
        return self.getDefaultOperatorSet()

    def _parse(self, method, input, operators):
        if operators is None:
            return method(input)
        else:
            return method(input, operators)

    def parse(self, input, operators=None):
        return self.parse_term(input, operators)

    def parse_term(self, input, operators=None):
        return self._parse(self.parseTerm, input, operators)

    def parse_struct(self, input, operators=None):
        return self._parse(self.parseStruct, input, operators)

    def parse_constant(self, input, operators=None):
        return self._parse(self.parseConstant, input, operators)

    def parse_var(self, input, operators=None):
        return self._parse(self.parseVar, input, operators)

    def parse_atom(self, input, operators=None):
        return self._parse(self.parseAtom, input, operators)

    def parse_numeric(self, input, operators=None):
        return self._parse(self.parseNumeric, input, operators)

    def parse_integer(self, input, operators=None):
        return self._parse(self.parseInteger, input, operators)

    def parse_real(self, input, operators=None):
        return self._parse(self.parseReal, input, operators)

    def parse_clause(self, input, operators=None):
        return self._parse(self.parseClause, input, operators)


@jpype.JImplementationFor("it.unibo.tuprolog.core.parsing.TermReader")
class _KtTermReader:
    def __jclass_init__(cls):
        pass

    @property
    def default_operator_set(self):
        return self.getDefaultOperatorSet()

    def _read(self, method, input, operators):
        input_stream = ensure_input_steam(input)
        if operators is None:
            return method(input_stream)
        else:
            return method(input_stream, operators)

    def read(self, input, operators=None):
        return self.read_term(input, operators)

    def read_term(self, input, operators=None):
        return self._read(self.readTerm, input, operators)

    def read_all(self, input, operators=None):
        return self.read_terms(input, operators)

    def read_terms(self, input, operators=None):
        return self._read(self.readTerms, input, operators)


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.core.parsing.*")



from tuprolog import logger
import jpype


@jpype.JImplementationFor("it.unibo.tuprolog.core.operators.Operator")
class _KtOperator:
    def __jclass_init__(cls):
        pass

    @property
    def functor(self):
        return self.getFunctor()

    @property
    def specifier(self):
        return self.getSpecifier()

    @property
    def priority(self):
        return self.getPriority()


@jpype.JImplementationFor("it.unibo.tuprolog.core.operators.OperatorSet")
class _KtOperatorSet:
    def __jclass_init__(cls):
        pass

    def __add__(self, other):
        return self.plus(other)

    def __sub__(self, other):
        return self.minus(other)


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.core.operators.*")
